# stylize.praat -- Praat include file
# Pitch contour stylization. 
# This file is included (indirectly) by prosogram.praat. It isn't a stand-alone script.
# Author: Piet Mertens
# Last modification: 2015-02-19


procedure af label$ indexname$ comment$
   select 'af_object$'
   .nc = Get number of columns
   Insert column (index)... .nc+1
   if (not af_object_init)
      Remove column (index)... 1
      af_object_init = 1
   endif
   .nc = Get number of columns
   Set column label (index)... .nc 'label$'
   'af_object$'_colname_index$ [.nc] = "'label$'_'.nc'"
   'indexname$' = .nc
endproc


procedure create_table_of_nuclei
   .nrows = nrof_nuclei_analysed
   if (.nrows == 0)
      call error_msg No nuclei were found. 
      .nrows = 1		; avoid crash
   endif
   nucldat_ncols = 1	; dummy: at least one column needed at creation of TableOfReal
   nucldatID = Create TableOfReal... nucl_data .nrows nucldat_ncols
   nucldat_available = 1
   af_object$ = "nucldatID"
   af_object_init = 0
   call af "nucl_t1"	"nucleusstarttime"	starttime of nucleus
   call af "nucl_t2"	"nucleusendtime"	endtime of nucleus
   call af "f0_min"	"f0_min"		f0 min (Hz) within nucleus            before stylization
   call af "f0_max"	"f0_max"		f0 max (Hz) within nucleus            before stylization
   call af "f0_median"	"f0_median"		f0 median (Hz) within nucleus         before stylization
   call af "f0_mean"	"f0_mean"		f0 mean (ST) within nucleus           before stylization
   call af "f0_start"	"f0_start"		f0 value (Hz) at start of nucleus     after stylization
   call af "f0_end"	"f0_end"		f0 value (Hz) at end of nucleus       after stylization
   call af "lopitch"	"lopitch"		f0 min (Hz) within nucleus            after stylization
   call af "hipitch"	"hipitch"		f0 max (Hz) within nucleus            after stylization
   call af "dynamic"	"dynamic"		0 = static, 1 = rising, -1 = falling
   call af "intrasyllab" "intrasyllab"		sum of pitch interval (ST) of tonal segments in nucleus (rises and falls compensate)
   call af "intersyllab" "intersyllab"		intersyllabic interval (ST) between end of previous nucleus and start of current one
   call af "up"		"intrasyllabup"		sum of upward pitch interval (ST) of tonal segments in nucleus	before stylization
   call af "down"	"intrasyllabdown"	sum of downward pitch interval (ST) of tonal segments in nucleus	before stylization
   call af "trajectory"	"trajectory"		sum of absolute pitch interval (ST) of tonal segments in nucleus (rises and falls add up)
   call af "f0_discont" "j_f0_discont"		f0 contains discontinuity
   call af "prnp_start" "j_prnp_start"		pitch-range normalized pitch at start of nucleus	after stylization
   call af "prnp_end"	"j_prnp_end"		pitch-range normalized pitch at end of nucleus	after stylization
   call af "prnp_intra"	"j_prnp_intra"		pitch-range normalized pitch of intranucleus variation after stylization
   call af "nucl_dur"	"nucleusdur"		nucleus duration
   call af "syll_dur"	"j_syllabledur"		syllable duration (only for appropriate segmentation methods)
   call af "vowel_dur"	"j_voweldur"		vowel duration (only for appropriate segmentation methods)
   call af "rhyme_dur"	"j_rhymedur"		rhyme duration (only for appropriate segmentation methods)
   call af "gap_left"	"internucleusdur"	time between end of previous nucleus and start of current one
   call af "loudness"	"loudness"		loudness peak in nucleus (only if parameter available)
   call af "int_peak"	"j_int_peak"		peak intensity in nucleus
if (use_duration_model)
   call af "en_sylldur" "j_en_sylldur"		elasticity normalized syllable duration (only for appropriate segmentation methods)
   call af "en_rhymedur" "j_en_rhymedur"	elasticity normalized rhyme duration (only for appropriate segmentation methods)
endif
if (calc_prominence)
   call af "promL2R1D_f0_mean" "promL2R1D_f0_mean"	prominence of mean pitch (Hz) wrt dynamic left/right context of 2+1 units
   call af "promL2R1D_f0_max" "promL2R1D_f0_max"	prominence of max pitch (Hz) wrt dynamic left/right context of 2+1 units
   call af "promL2R1D_nucldur" "j_promL2R1D_nucldur"	prominence of nucleus duration wrt dynamic left/right context of 2+1 units
   call af "promL2R1D_sylldur" "j_promL2R1D_sylldur"	prominence of syllable duration wrt dynamic left/right context of 2+1 units
   call af "promL2R1D_rhymedur"	"j_promL2R1D_rhymedur"	prominence of rhyme duration wrt dynamic left/right context of 2+1 units
   call af "endtime_syll" "endtime_syll"	endtime of syllable; used by plot_salience (only for appropriate segmentation method)
endif
if (use_duration_model and calc_prominence)
   call af "promL2R1D_en_sylldur" "j_promL2R1D_en_sylldur"	-
   call af "promL2R1D_en_rhymedur" "j_promL2R1D_en_rhymedur"	-
endif
   call af "hesitation"	"j_hesitation"		hesitation (required!)
   call af "speaker_id"	"speaker_id"		speaker ID number, from tier "speaker" in annotation file
   call af "before_pause" "j_before_pause"	syllable is followed by pause
 ; call af "local_rate_nucl" "j_localrate_nucl"	local speech rate in complete nuclei per second
endproc


procedure create_table_global_report
# Create variables for global report, showing results for all input files processed in one batch
    reportID = Create TableOfReal... global_report 1 1
    af_object$ = "reportID"
    af_object_init = 0

    reportID_empty = 1
    call af "SpeechTime" "j_speech_time"	
    call af "propphonation" "propphonation"	
    call af "proppause" "proppause"
    call af "SpeechRate" "j_speech_rate"	
    call af "PitchMeanOfST" "j_pitch_mean_of_ST"	
    call af "PitchStdevOfST" "j_pitch_stdev_of_ST"	
    call af "PitchRange" "j_pitch_range"	
    call af "PitchTop" "j_pitch_top"	
    call af "PitchBottom" "j_pitch_bottom"	
    call af "PitchMedian" "j_pitch_median"	
    call af "PitchMean" "j_pitch_mean"	
    call af "rf0_p02" "j_rf0_p02"	
    call af "rf0_p50" "j_rf0_p50"	
    call af "rf0_p98" "j_rf0_p98"	
    call af "rf0_mean" "j_rf0_mean"	
    call af "Gliss" "j_prop_gliss"	
    call af "Rises" "j_prop_rises"	
    call af "Falls" "j_prop_falls"	
    call af "TrajIntra" "j_rate_traj_intra"	
    call af "TrajInter" "j_rate_traj_inter"	
    call af "TrajPhon" "j_rate_traj_phon"	
    call af "TrajIntraZ" "j_rate_traj_intra_z"	
    call af "TrajInterZ" "j_rate_traj_inter_z"	
    call af "TrajPhonZ" "j_rate_traj_phon_z"	
    call af "nPVI" "j_nPVI"	
endproc


procedure create_table_global_report_OLD
# Create variables for global report, showing results for all input files processed in one batch
    j_speech_time = 1
    propphonation = 2
    proppause = 3
    j_speech_rate = 4
    j_pitch_mean_of_ST = 5
    j_pitch_stdev_of_ST = 6
    j_pitch_range = 7
    j_pitch_top = 8
    j_pitch_bottom = 9
    j_pitch_median = 10
    j_pitch_mean = 11
    j_prop_gliss = 12
    j_prop_rises = 13
    j_prop_falls = 14
    j_rate_traj_intra = 15
    j_rate_traj_inter = 16
    j_rate_traj_phon = 17
    j_rate_traj_intra_z = 18
    j_rate_traj_inter_z = 19
    j_rate_traj_phon_z = 20
    j_nPVI = 21
    f0_qu02 = 22
    f0_qu05 = 23
    f0_qu50 = 24
    f0_qu95 = 25
    f0_qu98 = 26
    sp_qu02 = 27
    sp_qu05 = 28
    sp_qu50 = 29
    sp_qu95 = 30
    sp_qu98 = 31

    reportID = Create TableOfReal... report 1 31
    reportID_empty = 1
    Set column label (index)... j_speech_time SpeechTime
    Set column label (index)... propphonation propphonation 
    Set column label (index)... proppause proppause 
    Set column label (index)... j_speech_rate SpeechRate 
    ;Set column label (index)... meannucldur meannucldur 
    Set column label (index)... j_pitch_mean_of_ST PitchMeanOfST 
    Set column label (index)... j_pitch_stdev_of_ST PitchStdevOfST
    Set column label (index)... j_pitch_range PitchRange 
    Set column label (index)... j_pitch_top PitchTop
    Set column label (index)... j_pitch_bottom PitchBottom
    Set column label (index)... j_pitch_median PitchMedian 
    Set column label (index)... j_pitch_mean PitchMean 
    Set column label (index)... j_prop_gliss Gliss 
    Set column label (index)... j_prop_rises Rises 
    Set column label (index)... j_prop_falls Falls 
    Set column label (index)... j_rate_traj_intra TrajIntra 
    Set column label (index)... j_rate_traj_inter TrajInter
    Set column label (index)... j_rate_traj_phon TrajPhon
    Set column label (index)... j_rate_traj_intra_z TrajIntraZ 
    Set column label (index)... j_rate_traj_inter_z TrajInterZ
    Set column label (index)... j_rate_traj_phon_z TrajPhonZ
    Set column label (index)... j_nPVI nPVI
    Set column label (index)... f0_qu02 f0_qu02
    Set column label (index)... f0_qu05 f0_qu05
    Set column label (index)... f0_qu50 f0_qu50
    Set column label (index)... f0_qu95 f0_qu95
    Set column label (index)... f0_qu98 f0_qu98
    Set column label (index)... sp_qu02 sp_qu02
    Set column label (index)... sp_qu05 sp_qu05
    Set column label (index)... sp_qu50 sp_qu50
    Set column label (index)... sp_qu95 sp_qu95
    Set column label (index)... sp_qu98 sp_qu98
endproc


procedure store_features .filecounter .long .outfname$
   if (.filecounter = 1)
      deleteFile(.outfname$)
   endif
   select nucldatID
   .nrows = Get number of rows
   if (not .long)
      Write to headerless spreadsheet file... '.outfname$'
   elsif (not segfile_available)
      Write to headerless spreadsheet file... '.outfname$'
   else		; long format
   # Store table in a temporary file
      call fname_parts '.outfname$'
      .tmpfile$ = result4$ + "_tmp_.txt"
      Write to headerless spreadsheet file... '.tmpfile$'
   # Find tier information
      .grid = nucleiID
      call tier_get .grid "^phon" l_phon_tier "No phon tier" 0
      call tier_get .grid "^syll" l_syll_tier "No syll tier" 0
      call tier_get .grid "^contour" l_contour_tier "No contour tier" 0
      call tier_get .grid "^prom" l_prom_tier "No prom tier" 0
      text$ = readFile$(.tmpfile$)
      deleteFile(.tmpfile$)
   # Read tmp file and append data, adding some categorical data not available inside TableOfReal
      .country$ = "NA"
      .discourse$ = "NA"
      # Handle first line of file, which contains header
         call next_line "text" line
      if (.filecounter = 1)
         line$ = replace_regex$ (line$, "\n$", "", 1)
         line$ = line$ + " syll rhyme prom contour pitchlevel corpus"
         if (corpus$ = "cprom")
            line$ = line$ + " discourse country"
         endif
         line$ = replace_regex$(line$, " ", "'tab$'", 0)
         appendFileLine(.outfname$, line$)
      endif
      # For all other lines: add categorical data columns
      .syl$ = "NA"
      .rhyme$ = "NA"
      .contour$ = "NA"
      .prom$ = "NA"
      .pitchlevel$ = "NA"
      if (corpus$ = "cprom")
         .discourse$ = replace_regex$(basename$, "\-\w*$", "", 1)
         .country$ = replace_regex$(basename$, "^\w*\-", "", 1)
      endif
      for .j to .nrows
         call next_line "text" line
         line$ = replace_regex$ (line$, "\n$", "", 1)
         select nucldatID
         .t1n = Get value... .j nucleusstarttime
         .t2n = Get value... .j nucleusendtime
         .midt = .t1n+(.t2n-.t1n)/2
         select .grid
         if (l_syll_tier)
            .k = Get interval at time... l_syll_tier .midt
            .syl$ = Get label of interval... l_syll_tier .k
            .t1s = Get start point... l_syll_tier .k
            .t2s = Get end point... l_syll_tier .k
            if (l_phon_tier)
               .t = .t1s		; start time of syllable
               .rhyme$ = ""		; sequence of sounds in rhyme
               .t2r = .t2s		; end time of rhyme
               .pos = -1		; assume we start in onset of syllable
               repeat			; concatenate the sequence of sounds in the syllable; obtain label of rhyme
                  .k = Get interval at time... l_phon_tier .t
                  .phon$ = Get label of interval... l_phon_tier .k
                  call is_syllabic '.phon$'
                  if (result)
                     .pos = 0			; in syllable peak
                     .t1r = Get start point... l_phon_tier .k
                  endif
                  if (.pos >= 0)
                     .rhyme$ = .rhyme$ + .phon$
                  endif
                  .t = Get end point... l_phon_tier .k
               until (.t >= .t2s)
            endif
         endif
         if (l_prom_tier)
            .k = Get interval at time... l_prom_tier .midt
            .prom$ = Get label of interval... l_prom_tier .k
         endif
         if (l_contour_tier)
            .k = Get interval at time... l_contour_tier .midt
            .contour$ = Get label of interval... l_contour_tier .k
            .pitchlevel$ = .contour$
            .pitchlevel$ = replace_regex$ (.pitchlevel$, "[rRfFSC_]", "", 0) 
            if (length (.pitchlevel$) < 1)
               .pitchlevel$ = "NA"
            endif
            .contour$ = replace_regex$ (.contour$, "[BLMHT]", "", 0) 
            if (index (.contour$, "C"))
               .contour$ = "NA"
            endif
            if (length (.contour$) < 1)
               .contour$ = "_"
            endif
         endif
         line$ = line$ + " '.syl$' '.rhyme$' '.prom$' '.contour$' '.pitchlevel$'"
         if (corpus$ = "cprom")
            line$ = line$ + " 'corpus$' '.discourse$' '.country$'"
         endif
         line$ = replace_regex$(line$, " ", "'tab$'", 0)
         appendFileLine(.outfname$, line$)
      endfor
   endif
if (0)
; Remove columns from right to left!
   Remove column (index)... promL2R1D_f0_max
   Remove column (index)... j_localrate_nucl
   Remove column (index)... endtime_syll
   Remove column (index)... j_before_pause
   Remove column (index)... promL2R1D_f0_mean
   Remove column (index)... loudness
   Remove column (index)... intrasyllab
endif
endproc


procedure safe_nuclei start_time_n end_time_n
# Adjust nuclei boundaries such that they are all voiced and pitch is defined
# Return in <result> the number of valid nuclei in analysis interval
   mindur_syl = 0.01	; minimum duration for syllable, otherwise skipped
   call interval_from_time nucleiID nucleus_tier start_time_n first_interval
   select nucleiID
   iNucl = first_interval
   prev_boundary = Get start point... safe_tier 1
   x1 = Get start point... nucleus_tier iNucl
   repeat
      select nucleiID
      iNucl = Get interval at time... nucleus_tier x1	; nrof intervals may change during process
      x1 = Get start point... nucleus_tier iNucl
      x2 = Get end point... nucleus_tier iNucl
      label$ = Get label of interval... nucleus_tier iNucl
      if (label$ = "a") ; "a" indicates syllabic nucleus, i.e. interval on which to apply stylization
         call defined_intersection pitchID x1 x2
         cx1 = result1
         cx2 = result2
;printline safe_nuclei: after defined_intersection x1='x1:4' x2='x2:4' cx1='cx1:4' cx2='cx2:4' result='result'
         select nucleiID
         if (result == 0)	; nucleus fully undefined
            Set interval text... nucleus_tier iNucl undef
         else
            if (cx1 > x1)	; undefined section at start of nucleus
               Insert boundary... nucleus_tier cx1
               Set interval text... nucleus_tier iNucl xL
               Set interval text... nucleus_tier iNucl+1 a
               iNucl += 1
            endif
            call octavejump nucleiID pitchID cx1 cx2
            cx1 = result1	; update safe interval
            cx2 = result2	; idem
;printline safe_nuclei: after octavejump cx1='cx1:4' cx2='cx2:4' result='result'
            select nucleiID
            if (result = 0)		; interval too short
               Set interval text... nucleus_tier iNucl short
            elsif (result3 == 1)	; discontinuity found
;printline safe_nuclei: f0_discont found at 'cx1:4'
               Insert boundary... nucleus_tier cx2
               Set interval text... nucleus_tier iNucl a
               Set interval text... nucleus_tier iNucl+1 skip
            elsif (cx2-cx1 < mindur_syl)
               Set interval text... nucleus_tier iNucl short
            elsif (cx2 < x2)
               Insert boundary... nucleus_tier cx2
               j = Get interval at time... nucleus_tier cx2
               Set interval text... nucleus_tier j xR
               Set interval text... nucleus_tier j-1 a
            endif
            if (cx2-cx1 >= mindur_syl)		; safe_tier
               if (cx1 > prev_boundary)
                  Insert boundary... safe_tier cx1
               endif
               Insert boundary... safe_tier cx2
               prev_boundary = cx2
               j = Get interval at time... safe_tier cx1+(cx2-cx1)/2
               Set interval text... safe_tier j a
            endif
         endif
      endif ; label="a"
      x1 = x2
; next line required because of rounding errors !
      intervals = Get number of intervals... nucleus_tier
   until (x2 >= end_time_n or iNucl = intervals)

   select nucleiID
   nn = Count labels... nucleus_tier a
   result = nn			; number of nuclei for analysis
endproc


procedure stylize_nuclei start_time_n end_time_n
   tier = nucleus_tier
   mindur_syl = 0.01	; minimum duration for syllable, otherwise skipped
   call interval_from_time nucleiID tier start_time_n first_interval
   call interval_from_time nucleiID tier end_time_n last_interval

   select nucleiID
   nrofnuclei = Get number of intervals... tier
   if (not reuse_nucl)
      call copy_tier nucleiID nucleus_tier nucleiID pointer_tier
      call tier_clear_text nucleiID pointer_tier
   endif

   ; fill some columns in table nucldatID 
   nucleus_ctr = 0		; counter for row
   for iNucl from first_interval to last_interval
      select nucleiID
      label$ = Get label of interval... tier iNucl
      if (label$ = "a") ; "a" indicates syllabic nucleus, i.e. interval on which to apply stylization
         x1 = Get start point... tier iNucl
         x2 = Get end point... tier iNucl
         nucleus_ctr += 1
         select nucleiID
         Set interval text... pointer_tier iNucl 'nucleus_ctr'
         select nucldatID
         Set row label (index)... nucleus_ctr 'x1:3'
         s$ = fixed$(x1,3)
         Set value... nucleus_ctr nucleusstarttime 's$'
         s$ = fixed$(x2,3)
         Set value... nucleus_ctr nucleusendtime 's$'
         s$ = fixed$(x2-x1,5)
         Set value... nucleus_ctr nucleusdur 's$'
         Set value... nucleus_ctr j_voweldur 0
         Set value... nucleus_ctr j_syllabledur 0
         Set value... nucleus_ctr j_rhymedur 0
         if (use_duration_model)
            Set value... nucleus_ctr j_en_sylldur 0
            Set value... nucleus_ctr j_en_rhymedur 0
         endif
      endif
   endfor
   ; calculate gap between successive nuclei to locate pauses
   select nucldatID
   mindur_pause_gap = 0.35		; min duration for gap between nuclei for pause 
   for j to nrof_nuclei_analysed
      x2 = Get value... j nucleusendtime
      if (j < nrof_nuclei_analysed)
         x1 = Get value... j+1 nucleusstarttime
      else
         x1 = x2 + 1
      endif
      if (x1-x2 >= mindur_pause_gap)
         Set value... j j_before_pause 1
      else
         Set value... j j_before_pause 0
      endif
   endfor

   prev_nucleus = 0	; index (into textgrid tier) of previous nucleus; 0 indicates "not found"
   for iNucl from first_interval to last_interval
      select nucleiID
      label$ = Get label of interval... tier iNucl
      if (label$ = "a") ; "a" indicates nucleus, i.e. interval to apply stylization on
         x1 = Get start point... tier iNucl
         x2 = Get end point... tier iNucl
         call stylize_nucleus iNucl x1 x2
      endif
   endfor

   ; call calc_localrate nucldatID j_localrate_nucl 2.5 2.5
   if (phones_available)	; calculate vowel duration for spreadsheet
      call calc_vowel_duration nucldatID j_voweldur
      call calc_nPVI nucldatID j_voweldur
   endif
   if (syllables_available and phones_available and segm_type <> segm_asyll) ; calculate syllable duration for spreadsheet
      call calc_rime_duration nucldatID j_rhymedur
      if (use_duration_model and fileReadable (duration_model_filename$))
         call calc_norm_duration "syllable" nucldatID j_en_sylldur 'duration_model_filename$'
         call calc_norm_duration "rhyme" nucldatID j_en_rhymedur 'duration_model_filename$'
      endif
   endif
endproc


procedure defined_intersection paramID t1 t2
# Find region within <t1>..<t2> for which parameter is defined
# Returns <result> = 0 when fully undefined
# Returns <result1> and <result2>, the defined interval
   result = 1
   result1 = t1
   result2 = t2
   select paramID
   dx = Get time step
   ok = 0
   t = t1
   while (ok == 0 and t <= t2)		; go right while undefined
      y = Get value at time... t Hertz Linear
      if (y == undefined)
         t += dx 
      else
         ok = 1
      endif
      result1 = t
   endwhile
   if (t > t2)
      result = 0
   else
      while (ok == 1 and t <= t2)	; go right while defined
         y = Get value at time... t Hertz Linear
         if (y == undefined)
            ok = 0
         else
            result2 = t
            t += dx 
         endif
      endwhile
      if (result1 >= t2)
         result = 0
      endif
   endif
endproc


procedure octavejump nuclID paramID t1 t2
# Find region within <t1>..<t2> for which pitch does not present discontinuities such as octave jumps
# Returns result = 0 when left with too short segment
# Returns result = 1 when OK
# Returns result1 and result2, the safe interval
# Returns result3 = 1 when discontinuity found, in which case <result2> is end of safe interval
# Stores position of discontinuity in (point) tier <discontinuity_tier> of <nuclID>
   result3 = 0			; no discontinuity found
   select paramID
   dx = Get time step
   ok = 1
   t = t1
   f1 = Get value at time... t Hertz Linear
   while (ok == 1 and t <= t2)
      select paramID
      f2 = Get value at time... t Hertz Linear
      if (abs(f2-f1)/min(f1,f2) > 0.3)	; was 0.2 initially and 0.5 in v2.7g
         ok = 0
         result3 = 1		; discontinuity found
         select nuclID
         Insert point... discontinuity_tier t
      else
         result2 = t
         t += dx
         f1 = f2
      endif
   endwhile
   result = 1
   if (result2 - t1 <= dx)
      result = 0
   endif
   result1 = t1
endproc


procedure stylize_nucleus iNucl x1 x2
# <x1>..<x2>	times of interval to stylize

   xmid = x1 + (x2-x1)/2
   select nucleiID
   i_ = Get interval at time... pointer_tier xmid
   s$ = Get label of interval... pointer_tier i_
   nucleus_ctr = 's$'		; index of row in table nucldatID
   select nucldatID
   pause_follows = Get value... nucleus_ctr j_before_pause
   glissando_local = glissando
   if (adaptive_glissando and pause_follows)
      glissando_local = glissando_low
   endif

# Step 1. Find turning points (TP) in contour, by order of importance.
# A TP is the point where the distance between the actual F0 and the linear fit of F0 values is largest.
# A TP is kept only if the difference in slope between the parts before and after 
# the TP exceeds the differential glissando threshold, and if at least 1 of the parts is an audible pitch movement.
# When a TP is found, additional TPs are searched for in the left part, until none are found.
# Then the search continues for the interval between the last TP and the end of the nucleus interval.
   nrofts = 1				; number of tonal segments
   select stylID
   Add point... x1 1
   Add point... x2 1
   i1 = Get nearest index from time... x1
   xL = x1				; xL..xR is time window under analysis
   xR = x2
;printline stylize nucleus='iNucl' start x1='x1:3' x2='x2:3' mindur_ts='mindur_ts:3'
   repeat
;printline stylize nucleus='iNucl' repeat1 xL='xL:3' xR='xR:3' mindur_ts='mindur_ts:3'
      nrofsplit = 0			; nrof turning points found in repeat loop 
      repeat				; find turning points
;printline stylize nucleus='iNucl' start xL='xL:3' xR='xR:3'
         split = 0			; nrof times split at turning point
         call is_audible "aud_A" xL xR
;printline stylize-2 xL='xL:3' xR='xR:3' AUD='aud_A' nrofts='nrofts'
         if (aud_A)
            call turning_point xL xR
;printline stylize-3 maxdiftime='maxdiftime'
            if (maxdiftime >= 0)	; found turning point
;d1 = maxdiftime - xL
;d2 = xR - maxdiftime
;printline stylize-4 leftpart='d1:3' rightpart='d2:3'
               if ((maxdiftime - xL >= mindur_ts) and (xR - maxdiftime >= mindur_ts))
                  call is_audible "aud_L" xL maxdiftime
                  g1 = slopeSTs		; slope (ST/s) for part left of turning point
                  call is_audible "aud_R" maxdiftime xR
                  g2 = slopeSTs		; slope (ST/s) for part right of turning point
;printline AUD_L='aud_L' slope='g1:2' AUD_R='aud_R' slope='g2:2'
                  if ((abs (g2-g1) > diffgt) and (aud_L or aud_R) )  
                     split = 1		; found a valid turning point 
                  endif
               endif
            endif ; (maxdiftime >= 0)
;printline stylize-5 nucleus='iNucl' split='split'
            if (split)	       
;printline stylize-5b nucleus='iNucl' split at t='maxdiftime:3'
               select stylID
               Add point... maxdiftime 1
               xR = maxdiftime
               nrofsplit += 1		; turning points inserted in this loop
               nrofts += 1		; additional tonal segment found
            endif
         endif
      until (split = 0)
      if (xR < x2)			; interval was split; continue segmentation for right side
         i1 += 1			; adjust xL..xR analysis window
         select stylID
         xL = Get time from index... i1
         xR = Get time from index... i1+1
;printline stylize-6 continue from t='xL:3'
      else				; no split...
         nrofsplit = 0			; prepare for end of repeat loop
         xL = x2
      endif
   until (nrofsplit == 0 and xL >= x2)

# Step 2. Actual stylisation
   cum_intra = 0		; cumulated intrasyllabic pitch variation
   cum_intra_up = 0		; sum of intrasyllabic pitch rises
   cum_intra_down = 0		; sum of intrasyllabic pitch falls
   cum_abs_intra = 0		; cumulated absolute intrasyllabic pitch variation
   dynamic_type = 0		; dynamic_type, 0 = static, 1 = rising, -1 = falling
   dynamic_up = 0
   dynamic_down = 0
   select stylID
   i = Get nearest index from time... x1
   i2 = Get nearest index from time... x2
   ts = 1			; index of tonal segment under analysis
   while (i < i2)		; for each tonal segment
      select stylID
      xL = Get time from index... i
      xR = Get time from index... i+1
      call is_audible "aud_A" xL xR
;printline stylize_nucleus, step 2: ts='ts' ('xL:3'..'xR:3'), aud_A='aud_A' intST='dist:2'
      intST = dist		; pitch interval (in ST) in current tonal segment
      cum_intra += intST
      cum_abs_intra += abs (intST)
      cum_intra_up += max (intST, 0)
      cum_intra_down += min (intST, 0)
      # Check special case of two inaudible parts. e.g. bell-shaped contour
      if (aud_A = 1 and nrofts = 1)
;printline Check special case xL='xL:3' xR='xR:3'
            call turning_point xL xR
;printline stylize_nucleus, step 2, special case: maxdiftime='maxdiftime:3'
            if (maxdiftime >= 0)	; turning point found
               call is_audible "aud_L" xL maxdiftime
               g1 = slopeSTs
               call is_audible "aud_R" maxdiftime xR
               g2 = slopeSTs
               d1 = maxdiftime - xL
               d2 = xR - maxdiftime
;printline stylize_nucleus, step 2, special case: AUD_L='aud_L' slope='g1:2' AUD_R='aud_R' slope='g2:2'
               if (aud_L == 0 and aud_R == 0)  
                  if ((g1 > 0 and g2 < 0) or (g1 < 0 and g2 > 0))  
                     aud_A = 0	; consider inaudible
                     intST = 0
                  endif
               endif
            endif
      endif ; special case
      select pitchID
      yR = Get value at time... xR Hertz Linear
      yL = Get value at time... xL Hertz Linear
      yM = Get quantile... xL xR 0.5 Hertz
;printline xL='xL:3' yL='yL:3' xR='xR:3' yR='yR:3' yM='yM:3'
      select stylID
      if (ts = 1)		; first tonal segment of nucleus => also set value at start
         if (aud_A = 0)
            yR = yM		; normalize pitch to modal pitch
            yL = yR
         endif
         Remove point... i	; to replace value of point at xL
         Add point... xL yL	; set Y value of turning point at xL
         pv_lo = min (yL, yR)
         pv_hi = max (yL, yR)
         pv_start = yL
      endif
      Remove point... i+1	; to replace value of point at xR
      Add point... xR yR	; set Y value of turning point at xR
      if (aud_A)
         distST = 12 * log2 (yR/yL)
         if (distST > 0)
            dynamic_up += intST
         else
            dynamic_down += intST
         endif
         dynamic_type = 1
      endif
      pv_lo = min (pv_lo, yR)
      pv_hi = max (pv_hi, yR)
      ts += 1
      i += 1
   endwhile
   if (dynamic_type == 1)
      if (abs(dynamic_down) > dynamic_up)
         dynamic_type = -1
      endif
   endif
   select pitchID
   v_f0_min = Get minimum... x1 x2 Hertz Parabolic
   v_f0_max = Get maximum... x1 x2 Hertz Parabolic
   pv_median = Get quantile... x1 x2 0.50 Hertz
   pv_mean = Get mean... x1 x2 semitones re 1 Hz
   if (prev_nucleus > 0)	; already found a nucleus in left context
      call distST prev_f0_end pv_start
      select nucleiID
      prev_x2 = Get end point... nucleus_tier prev_nucleus
   else				; first nucleus in analysis window
      prev_x2 = anal_t1
      distST = 0
   endif
   if (syllables_available)
      select nucleiID
      imid = Get interval at time... syllable_tier xmid
      syllt1 = Get start point... syllable_tier imid
      syllt2 = Get end point... syllable_tier imid
      sylldur = syllt2 - syllt1
   else
      sylldur = undefined
      syllt2 = undefined
   endif

; Detect whether the original nucleus contained an F0 discontinuity, which was detected by octave_jump called by safe_nuclei
   select nucleiID
   k = Get number of points... discontinuity_tier
   i_ = Get high index from time... discontinuity_tier x1
   if (i_ > 0 and i_ <= k)		; time x1 > time of last discontinuity
      t_ = Get time of point... discontinuity_tier i_
   else
      t_ = -1 
   endif
   v_f0_discont = 0
   if (t_ >= x1-time_step and t_ <= x2+time_step)	; discontinuity within nucleus
      v_f0_discont = 1
   endif
   
; Store all parameters for syllable in table
   select nucldatID
   Set value... nucleus_ctr f0_start floor(pv_start)
   Set value... nucleus_ctr f0_end floor(yR)
   Set value... nucleus_ctr f0_min floor(v_f0_min)
   Set value... nucleus_ctr f0_max floor(v_f0_max)
   Set value... nucleus_ctr f0_median floor(pv_median)
   ; s$ = fixed$(pv_mean,3)
   Set value... nucleus_ctr f0_mean floor(pv_mean)
   Set value... nucleus_ctr lopitch floor(pv_lo)
   Set value... nucleus_ctr hipitch floor(pv_hi)
   Set value... nucleus_ctr dynamic dynamic_type
   s$ = fixed$(cum_intra,3)
   Set value... nucleus_ctr intrasyllab 's$'
   s$ = fixed$(cum_intra_up,3)
   Set value... nucleus_ctr intrasyllabup 's$'
   s$ = fixed$(cum_intra_down,3)
   Set value... nucleus_ctr intrasyllabdown 's$'
   s$ = fixed$(distST,3)
   Set value... nucleus_ctr intersyllab 's$'
   s$ = fixed$(cum_abs_intra,3)
   Set value... nucleus_ctr trajectory 's$'
   Set value... nucleus_ctr j_f0_discont v_f0_discont
; internucleusdur = time between end of previous nucleus and start of current one
   s$ = fixed$((x1 - prev_x2),3)
   Set value... nucleus_ctr internucleusdur 's$'
   if (syllables_available)
      s$ = fixed$(sylldur,4)
      Set value... nucleus_ctr j_syllabledur 's$'
   endif
;   endtime_syll: used by plot_salience (only for appropriate segmentation method)
   if (calc_prominence)
      Set value... nucleus_ctr endtime_syll syllt2
   endif
   if (include_loudness and loudness_available)
      select loudnessID
      v = Get maximum... x1 x2 None
      s$ = fixed$(v,3)
      select nucldatID
      Set value... nucleus_ctr loudness 's$'
   endif
   select intensityID
   v = Get maximum... x1 x2 None
   s$ = fixed$(v,1)
   select nucldatID
   Set value... nucleus_ctr j_int_peak 's$'
   prev_f0_end = yR			; save f0_end for next syllable
   prev_nucleus = iNucl			; used for next nucleus
   prev_nucleus_ctr = nucleus_ctr	; used for next nucleus
   last_x2 = x2				; used for last nucleus
endproc


procedure distST f1 f2
   distST = 12 * log2 (f2/f1)
endproc


procedure slopeSTs first last
# Calculate slope of F0 variation (in ST/s) in time interval <first>..<last>
# Return slope in variable 'slopeSTs'.
# Return pitch intervak (in ST) in variable 'dist'
   select pitchID
   max = Get maximum... first last Hertz None
   min = Get minimum... first last Hertz None
   tmax = Get time of maximum... first last Hertz None
   tmin = Get time of minimum... first last Hertz None
   if (tmin <= tmax) 
      dist = 12 * log2 (max/min)
   else 
      dist = 12 * log2 (min/max)
   endif
   slopeSTs = dist / (last-first)
endproc


procedure is_audible varname$ first last
# Evaluate audibility of pitch change for frequency values (in Hz)
# and store boolean result in variable with name 'varname$'
   duration = last-first
   call slopeSTs first last
   if (abs (slopeSTs) >= glissando_local/(duration*duration))
      'varname$' = 1
   else 
      'varname$' = 0
   endif
endproc


procedure turning_point first last
# Find most important turning point.
# returns <maxdiftime> = time of turning point in time interval <first>..<last>
# returns -1 if max difference too small ( < 1 ST ) 
   select pitchID
   dx = Get time step
   q1 = Get frame number from time... first
   f0_x1 = Get value in frame... q1 Hertz
   if (f0_x1 == undefined)
      x1_time = Get time from frame... q1
      call fatal_error turning_point: Pitch undefined at (left boundary) time='x1_time' first='first' last='last' q1='q1' q2='q2' 
   endif
   q2 = Get frame number from time... last
   f0_x2 = Get value in frame... q2 Hertz
   if (f0_x2 == undefined)
      x2_time = Get time from frame... q2
      call fatal_error turning_point: Pitch undefined at (right boundary) time='x2_time' first='first' last='last' q1='q1' q2='q2' 
   endif
   a = f0_x1
   b = (f0_x2 - f0_x1) / ((q2-q1)*dx)
   maxdif = 0
   maxdiffit = 1
   maxdiftime = first
   maxdifq = q1
   for q from q1 to q2
      fit = a + b * ((q-q1)*dx)
      f0_x = Get value in frame... q Hertz
      if (f0_x == undefined)
         time = Get time from frame... q
         call fatal_error turning_point: Pitch undefined at time='time:4' first='first' last='last' q1='q1' q2='q2' 
      endif
      dy = abs (f0_x - fit)
      if (dy > maxdif)
         maxdif = dy
         maxdifq = q
         maxdiffit = fit
      endif
   endfor
   if (maxdif = 0)
      maxdiftime = -1
   elsif (abs(12 * log2 (maxdif/maxdiffit)) < 1)	; smaller than 1 ST 
      maxdiftime = -1
   else
      maxdiftime = Get time from frame... maxdifq
   endif
endproc


procedure calc_vowel_duration table dst
# <dst>		column where results are stored
   select table
   rows = Get number of rows
   for row to rows	; for each nuclei in the signal
      s_$ = ""
      select table
      Set value... row dst 0
      x1 = Get value... row nucleusstarttime
      x2 = Get value... row nucleusendtime
      xmid = x1 + (x2-x1)/2
      select nucleiID
      if (syllables_available and phones_available)
         i_ = Get interval at time... syllable_tier xmid
         s_$ = Get label of interval... syllable_tier i_ 
         x1 = Get start point... syllable_tier i_
         x2 = Get end point... syllable_tier i_
         i = Get interval at time... phone_tier x1
         i2 = Get interval at time... phone_tier x2
         repeat
            select nucleiID
            label$ = Get label of interval... phone_tier i
            call is_vowel 'label$'
            if (is_vowel)
               x1 = Get start point... phone_tier i
               x2 = Get end point... phone_tier i
            elsif (i+1 == i2)
               call msg No vowel in syllable <'s_$'> at 'xmid:3'
               ; use syllable duration
            endif
            i += 1
         until (is_vowel or i == i2)
      elsif (phones_available)
         i = Get interval at time... phone_tier xmid
         label$ = Get label of interval... phone_tier i
         call is_vowel 'label$'
         if (is_vowel)
            x1 = Get start point... phone_tier i
            x2 = Get end point... phone_tier i
         endif
      endif
      select table
      s_$ = fixed$(x2-x1,4)
      Set value... row dst 's_$'
   endfor
endproc


procedure calc_rime_duration table dst
# Calculate duration of rhyme using intervals in the syllable and phoneme tiers
# <dst>		column where results are stored
   select table
   .rows = Get number of rows
   for j to .rows	; nrof_nuclei_analysed
      Set value... j j_rhymedur undefined	; prepare for possible error in annotation or lacking annotation
   endfor
   for j to nrof_nuclei_analysed		; calculate rime duration
      select table
      x1 = Get value... j nucleusstarttime
      x2 = Get value... j nucleusendtime
      ; use midtime to deal with truncating errors for nucleusstarttime and endtime
      call interval_from_time nucleiID syllable_tier (x1+(x2-x1)/2) syll
      syll_x1 = Get start point... syllable_tier syll
      syll_x2 = Get end point... syllable_tier syll
;if (x1 > 37.02 and x1 < 37.18)
;printline calc_rime_duration: x1='x1:6' x2='x2:6' syll_x1='syll_x1:6', syll_x2='syll_x2:6' 
;endif
      call interval_from_time nucleiID phone_tier syll_x1 ph1
      call interval_from_time nucleiID phone_tier syll_x2-0.001 ph2
      phon = ph1
      .nrof_syllabics = 0
      repeat
          select nucleiID
          label$ = Get label of interval... phone_tier phon
          call is_syllabic 'label$'
;if (x1 > 37.02 and x1 < 37.18)
;printline calc_rime_duration: syll_x1='syll_x1:4', syll_x2='syll_x2:4' 
;printline ph1='ph1' ph2='ph2' phon='phon' label=<'label$'> syllabic='result'
;endif
         if (result)
            phon_x1 = Get start point... phone_tier phon
            select table
           .s$ = fixed$(syll_x2-phon_x1, 4)
            Set value... j dst '.s$'
            .nrof_syllabics += 1
         endif
         phon += 1
      until (result or phon > ph2)
      if (.nrof_syllabics == 0)
         call msg calc_rime_duration: Syllable without syllabic sound at time 'syll_x1:3'
      endif
   endfor
endproc


procedure calc_nPVI table src
# Calculate Normalized Pairwise Variability Index on data in <table>
# <src>		column where values are taken from
# <result>	nPVI for all data in src
   select table
   rows = Get number of rows
   sum = 0
   if (rows > 1)		; need at least 2 syllables
      for row from 2 to rows	; for each nucleus in the signal
         y = Get value... row src
         y1 = Get value... row-1 src 
         sum += abs( (y1 - y) / ((y1 + y)/2) )
      endfor
      result = 100*sum/(rows-1)
   else
      result = 0
   endif
endproc


procedure prosodic_profile
 if (nrof_nuclei_analysed == 0)
   filedelete 'statsfile$'
   fileappend 'statsfile$' Prosodic profile for input file: 'signalfile$'
   fileappend 'statsfile$' The prosodic profile could not be calculated because no nuclei were detected.'newline$'
   call msg Prosodic profile not calculated: no nucleu detected in speech signal.
 elsif (not needs_pitchrange)
   filedelete 'statsfile$'
   fileappend 'statsfile$' Prosodic profile for input file: 'signalfile$'
   fileappend 'statsfile$' Prosodic profile calculation requires selection of full time range of speech signal. 'newline$'
   call msg Prosodic profile not calculated: requires selection of full time range of signal.
 else ; nrof_nuclei_analysed > 0 and pitchrange calculated
   min_pause_duration = 0.3
   select nucldatID
   nrof_nucl = Get number of rows
   sum_nucldur = 0
   sum_internucldur = 0	; time between nuclei, corrected for pauses
   sum_gapdur = 0	; time of pauses

   for j from 1 to nrof_nucl
      x1$ = Get row label... j
      sum_nucldur += Get value... j nucleusdur
      v_ind = Get value... j internucleusdur
      if (v_ind >= min_pause_duration)		; pause to LEFT of current nucleus
         sum_gapdur += v_ind
      else
         sum_internucldur += v_ind
      endif
   endfor
   total_dur      = anal_t2 - anal_t1
   phonation_time = sum_internucldur + sum_nucldur
   speech_time    = phonation_time + sum_gapdur
   prop_pause     = 100 * sum_gapdur / speech_time
   prop_phonation = 100 * phonation_time / speech_time
   speech_rate    = nrof_nucl / phonation_time

   date_$ = date$ ()
call msg Writing prosodic profile to: 'statsfile$'
   filedelete 'statsfile$'
   w$ = "fileappend ""'statsfile$'"" 'newline$'"
   'w$' Prosodic profile for input file: 'signalfile$'
   'w$'
   'w$' Date: 'date_$'
   'w$' Segmentation type: 'segmentation_name$'
   'w$' Time: (assumes there's only 1 speaker) 
   'w$'     total speech time        ='speech_time:2' s (= internucleus time + intranucleus time + pause time)
   'w$'     estimated phonation time ='phonation_time:2' ('prop_phonation:2'% of speech time) (= internucleus time + intranucleus time)
   'w$'     estimated pause time     ='sum_gapdur:2' ('prop_pause:2'% of speech time) (= when internucleus time >= 'min_pause_duration')
   'w$'     estimated speech rate    ='speech_rate:2' (nrof_nuclei/phonation_time)

   'w$' Nucleus: 'nrof_nucl' nuclei in signal
   if (nrof_nucl < 100)
      'w$' 'newline$' WARNING: The global measures below are only meaningful for speech samples of at least 100 nuclei (syllables).'newline$'
   endif
   if (nrof_nucl <= 1)
      'w$' 'newline$' ERROR: Insufficient number of syllables ('nrof_nucl') for calculation of measures.'newline$'
   endif
   select nucldatID
   mean_nucldur = Get column mean (label)... nucl_dur
   stdev_nucldur = Get column stdev (label)... nucl_dur
   if (nrof_nucl <= 1)
      stdev_nucldur = 0
   endif
   'w$' Duration: 
   'w$'     Nucleus duration: 
   'w$'        mean='mean_nucldur:3'(s) stdev='stdev_nucldur:3' summed nucleus duration='sum_nucldur:2'(s)
   call calc_nPVI nucldatID nucleusdur
   nPVI_nucldur = result
   'w$'     nPVI (nucleus duration)='result:0' (assumes there's only 1 speaker)
   if (syllables_available)
      call calc_nPVI nucldatID j_syllabledur
     'w$'     nPVI (syllable duration)='result:0' (assumes there's only 1 speaker)
   endif
   if (phones_available)
      call calc_vowel_duration nucldatID j_voweldur
      call calc_nPVI nucldatID j_voweldur
     'w$'     nPVI (vowel duration)='result:0' (assumes there's only 1 speaker)
   endif

   'w$'
   call pitchrange_speakers_report
   'w$' 'result$'

   call raw_pitchrange_speakers_report
   'w$' 'result$'

   call pitchprofile_speakers_report
   'w$' 'result$'

'w$' SpeechTime = total speech time (in s) = internucleus time + intranucleus time + pause time
'w$' PhonTime   = phonation time (in s) = without pauses = internucleus time + intranucleus time
'w$' PropPhon   = proportion (%) of estimated phonation time (= internucleus time + intranucleus time) to speech time
'w$' PropPause  = proportion (%) of estimated pause time (= when internucleus time >= 0.3) to speech time
'w$' SpeechRate = estimated speech rate (in syll/s) = nrof_nuclei/phonation_time
'w$' MeanOfST   = mean of pitch values, where values are min and max pitch in ST for each syllable
'w$' StdevOfST  = stdev of pitch values, where values are min and max pitch in ST for each syllable
'w$' PitchRange = estimated pitch range (in ST) (2%-98% percentiles of data in nuclei without discontinuities)
'w$' Gliss      = proportion (%) of syllables with large pitch movement (abs(distance) >= 4ST)
'w$' Rises      = proportion (%) of syllables with pitch rise (>= 4ST)  
'w$' Falls      = proportion (%) of syllables with pitch fall (<= -4ST)
'w$' NuclDur    = sum of durations for nuclei for this speaker
'w$' InterNuclDur = sum of durations between successive nuclei for this speaker
'w$' TrajIntra  = pitch trajectory (sum of absolute intervals) within syllabic nuclei, divided by duration (in ST/s) 
'w$' TrajInter  = pitch trajectory (sum of absolute intervals) between syllabic nuclei (except pauses or speaker turns), divided by duration (in ST/s) 
'w$' TrajPhon   = sum of TrajIntra and TrajInter, divided by phonation time (in ST/s) 
'w$' TrajIntraZ = as TrajIntra, but for pitch trajectory in standard deviation units on ST scale (z-score) (in sd/s) 
'w$' TrajInterZ = as TrajInter, but for pitch trajectory in standard deviation units on ST scale (z-score) (in sd/s)
'w$' TrajPhonZ  = as TrajPhon,  but for pitch trajectory in standard deviation units on ST scale (z-score) (in sd/s)
'w$'

      ;if (global_report)
      # store some values for the global profile spreadsheet (covering multiple speech files)
         select reportID
         for .speaker from 1 to speakers    ; nrofspeakers in current file
            .nrows = Get number of rows
            Insert row (index)... .nrows+1
            if (reportID_empty)
               Remove row (index)... 1
               reportID_empty = 0
            endif
            .row = Get number of rows
            label$ = basename$ + "_" + speaker_label'.speaker'$
            Set row label (index)... .row "'label$'"
            .d1 = extractNumber (speaker_profile_'.speaker'$, "NUCL_DUR=")
            .d2 = extractNumber (speaker_profile_'.speaker'$, "INTERNUCL_DUR=")
            .d3 = extractNumber (speaker_profile_'.speaker'$, "PAUSE_DUR=")
            v = .d1 + .d2 + .d3
            Set value... .row j_speech_time 'v:2'
            v = 100 * (.d1 + .d2) / (.d1 + .d2 + .d3)
            Set value... .row propphonation 'v:2'
            v = 100 * .d3 / (.d1 + .d2 + .d3)
            Set value... .row proppause 'v:2'
            v = extractNumber (speaker_profile_'.speaker'$, "SPEECH_RATE=")
            Set value... .row j_speech_rate 'v:2'
            v = extractNumber (speaker_range_'.speaker'$, "MEAN_OF_ST=")
            Set value... .row j_pitch_mean_of_ST 'v:1'
            v = extractNumber (speaker_range_'.speaker'$, "STDEV_OF_ST=")
            Set value... .row j_pitch_stdev_of_ST 'v:1'
            v = extractNumber (speaker_range_'.speaker'$, "RANGE_ST=")
            Set value... .row j_pitch_range 'v:1'
            v = extractNumber (speaker_range_'.speaker'$, "TOP_Hz=")
            Set value... .row j_pitch_top 'v:0'
            v = extractNumber (speaker_range_'.speaker'$, "BOTTOM_Hz=")
            Set value... .row j_pitch_bottom 'v:0'
            v = extractNumber (speaker_range_'.speaker'$, "MEDIAN_Hz=")
            Set value... .row j_pitch_median 'v:0'
            v = extractNumber (speaker_range_'.speaker'$, "MEAN_Hz=")
            Set value... .row j_pitch_mean 'v:0'
            Set value... .row j_nPVI 'nPVI_nucldur:2'
            v = extractNumber (speaker_range_'.speaker'$, "RF0_P02_Hz=")
            Set value... .row j_rf0_p02 'v:0'
            v = extractNumber (speaker_range_'.speaker'$, "RF0_P50_Hz=")
            Set value... .row j_rf0_p50 'v:0'
            v = extractNumber (speaker_range_'.speaker'$, "RF0_P98_Hz=")
            Set value... .row j_rf0_p98 'v:0'
            v = extractNumber (speaker_range_'.speaker'$, "RF0_MEAN_Hz=")
            Set value... .row j_rf0_mean 'v:0'
            v = extractNumber (speaker_profile_'.speaker'$, "GLISS=")
            Set value... .row j_prop_gliss 'v:2'
            v = extractNumber (speaker_profile_'.speaker'$, "RISES=")
            Set value... .row j_prop_rises 'v:1'
            v = extractNumber (speaker_profile_'.speaker'$, "FALLS=")
            Set value... .row j_prop_falls 'v:2'
            v = extractNumber (speaker_profile_'.speaker'$, "TRAJ_INTRA_RATE=")
            Set value... .row j_rate_traj_intra 'v:2'
            v = extractNumber (speaker_profile_'.speaker'$, "TRAJ_INTER_RATE=")
            Set value... .row j_rate_traj_inter 'v:2'
            v = extractNumber (speaker_profile_'.speaker'$, "TRAJ_PHON_RATE=")
            Set value... .row j_rate_traj_phon 'v:2'
            v = extractNumber (speaker_profile_'.speaker'$, "TRAJ_INTRA_Z_RATE=")
            Set value... .row j_rate_traj_intra_z 'v:2'
            v = extractNumber (speaker_profile_'.speaker'$, "TRAJ_INTER_Z_RATE=")
            Set value... .row j_rate_traj_inter_z 'v:2'
            v = extractNumber (speaker_profile_'.speaker'$, "TRAJ_PHON_Z_RATE=")
            Set value... .row j_rate_traj_phon_z 'v:2'
         endfor ; .speaker
      ;endif ; global_report
 endif ; nrof_nuclei_analysed
endproc


procedure pitchrange_speakers
; Compute pitch range for each speaker using high and low pitch values of each syllable.
; Should be called after stylization.
   if (speakers < 1)				; speaker are numbered from 1 to N
      call error_msg pitchrange_speakers: Expected >= 1 speaker
   endif
   if (needs_pitchrange < 1)
      call fatal_error pitchrange_speakers: no pitch range measurement activated
   endif
   select nucldatID
   rows = Get number of rows
   if (rows < 1)
      call fatal_error pitchrange_speakers: 0 rows in nucldatID 
   endif

   for speaker_j from 1 to speakers
      label$ = speaker_label'speaker_j'$
      speaker_nnucl [speaker_j] = 0	; nrof nuclei for this speaker

; call pitch_to_zscore pitchID 0

; For speaker X, compute quantiles of low and high pitch values of each nucleus after stylization. 
; To separate speaker X from others, a temporary table is used and data for other speakers are discarded from it.
; Also discard nuclei with a pitch discontinuity.
      select nucleiID
      k = Get number of points... discontinuity_tier
      select nucldatID
      rows = Get number of rows
      rows2 = rows * 2		; uses 2 pitch values per syllable
      tmptableID = Create Table without column names... pitchvalues rows2 4
      Set column label (index)... 1 pitch_Hz
      Set column label (index)... 2 pitch_ST
      Set column label (index)... 3 starttime
      Set column label (index)... 4 f0_Hz
      n = 0			; nrof data points
      for j from 1 to rows
         select nucldatID
         id = Get value... j speaker_id
;printline pitchrange_speakers: row='j' speaker_j='speaker_j' label='label$' id='id'
         if (id == speaker_j)
           speaker_nnucl [speaker_j] += 1
           t1_ = Get value... j nucleusstarttime
           t2_ = Get value... j nucleusendtime
           vlo = Get value... j lopitch
           vhi = Get value... j hipitch
           .f0_hi = Get value... j f0_max
           .f0_lo = Get value... j f0_min
           select nucleiID
           i_ = Get high index from time... discontinuity_tier t1_
           if (i_ > 0 and i_ <= k)		; time t1_ > time of last discontinuity
              t_ = Get time of point... discontinuity_tier i_
           else
              t_ = -1 
           endif
           if (t_ >= t1_-time_step and t_ <= t2_+time_step)	; avoid data at discontinuity
;printline pitchrange: discontinuity, skipped nucleus at 't1_:2'
           else
;printline pitchrange: nucleus at 't1_:2', lo='vlo:0' hi='vhi:0'
              select tmptableID
              Set numeric value... n+1 pitch_Hz vlo
              Set numeric value... n+1 pitch_ST hertzToSemitones (vlo) - hertzToSemitones(1)
              Set numeric value... n+1 starttime t1_
              Set numeric value... n+1 f0_Hz .f0_lo
              Set numeric value... n+2 pitch_Hz vhi
              Set numeric value... n+2 pitch_ST hertzToSemitones (vhi) - hertzToSemitones(1)
              Set numeric value... n+2 starttime t1_
              Set numeric value... n+2 f0_Hz .f0_hi
              n += 2
           endif
         endif
      endfor
;;printline pitchrange_speakers: speaker='speaker_j' ('label$') number of valid points n='n' 
      if (n = 0)		; no valid points are available for this speaker, use default values
         call msg Warning: pitchrange measurement: no pitch data for speaker 'speaker_j' "'label$'"  
         mean = 150
         median = 150
         bottom = 150
         top    = 150
         stdev  = 0
         mean_of_ST = 0
         stdev_of_ST = 0
      else
         if (n < 200)
            call msg Warning: too few syllables for robust pitchrange detection of speaker "'label$'"  
         endif
         select tmptableID
;;printline pitchrange_speakers: n='n' nrows2='nrows2' 
         if (n < rows2)		; number of row in table should match number of values
            row = rows2
            while (row > n)	; remove unused rows at end of table
               Remove row... row
               row -= 1
            endwhile
         endif
;;.nrows = Get number of rows
;;printline pitchrange_speakers: nrows='.nrows'
         medianST = Get quantile... pitch_ST 0.5
         ;mean = Get mean... pitch_Hz
         ;meanST = hertzToSemitones (mean) - hertzToSemitones(1)
         select tmptableID
         rows = Get number of rows
         row = rows
         while (row > 0)
            v = Get value... row pitch_ST
            if (abs(v - medianST) > 18)			; discard manifest errors
;t1_ = Get value... row starttime
;printline pitchrange: time='t1_:2', v='v:0' SKIPPED : abs distance from median > 18 ST
               Remove row... row
            endif
            row -= 1
         endwhile
         rows = Get number of rows
         mean = Get mean... pitch_Hz
         median = Get quantile... pitch_Hz 0.5
         bottom = Get quantile... pitch_Hz 0.02
         top    = Get quantile... pitch_Hz 0.98
         stdev  = Get standard deviation... pitch_Hz
         mean_of_ST = Get mean... pitch_ST
         stdev_of_ST = Get standard deviation... pitch_ST
         mean_rf0 = Get mean... f0_Hz
         p50_rf0 = Get quantile... f0_Hz 0.5
         p02_rf0 = Get quantile... f0_Hz 0.02
         p98_rf0 = Get quantile... f0_Hz 0.98
      endif
      if (stdev == undefined)		; when insufficient data
         stdev = 0
      endif
      if (stdev_of_ST == undefined)	; when insufficient data
         stdev_of_ST = 0
      endif
      meanST   = hertzToSemitones (mean) - hertzToSemitones(1)
      medianST = hertzToSemitones (median) - hertzToSemitones(1)
      bottomST = hertzToSemitones (bottom) - hertzToSemitones(1)
      topST    = hertzToSemitones (top) - hertzToSemitones(1)
      select tmptableID
      Remove
      n_ = rows/2			; nrof nuclei used 
      label$ = speaker_label'speaker_j'$
      range = topST - bottomST
      upper_range = 12 * log2 (top/median)
      lower_range = range - upper_range
      speaker_range_'speaker_j'$ = 
      ... "BOTTOM_Hz='bottom:0' BOTTOM_ST='bottomST:1' MEDIAN_Hz='median:0' MEDIAN_ST='medianST:1' " +
      ... "MEAN_Hz='mean:0' MEAN_ST='meanST:1' STDEV_HZ='stdev:2' MEAN_OF_ST='mean_of_ST:2' STDEV_OF_ST='stdev_of_ST:3' " +
      ... "TOP_Hz='top:0' TOP_ST='topST:1' " +
      ... "RANGE_ST='range:1' UPPER_RANGE_ST='upper_range:1' LOWER_RANGE_ST='lower_range:1' NROFNUCL='n_' " +
      ... "RF0_P02_Hz='p02_rf0:0' RF0_P98_Hz='p98_rf0:0' RF0_P50_Hz='p50_rf0:0' RF0_MEAN_Hz='mean_rf0:0'"

; For speaker X, compute some values using a temporary copy of nucldat, from which data from other speakers will be removed. 
; Computed values include mean and stdev for intrasyllabic and intersyllabic pitch variation,
      min_pause_duration = 0.3
      select nucldatID
      tmptableID = Copy... tmptable
      rows = Get number of rows
      for j from 1 to rows				; Modify value for <intersyllab> at speaker turn 
         t1_ = Get value... j nucleusstarttime
         spkr = Get value... j speaker_id
         if (j > 1)
            prev = Get value... j-1 speaker_id
            if (spkr == speaker_j and spkr <> prev)	; speaker has changed
               Set value... j intersyllab 0
            endif
         endif
      endfor
      j = 1
      while (j <= rows)					; Discard info from other speakers
         spkr = Get value... j speaker_id
         if (spkr <> speaker_j)
            Remove row (index)... j
            rows -= 1
            j -= 1
         endif
         j += 1
      endwhile
      sum_traj_intra = 0	; intrasyllabic trajectory
      sum_traj_inter = 0	; intersyllabic trajectory
      sum_nucldur = 0
      sum_internucldur = 0	; time between nuclei, corrected for pauses
      sum_pausedur = 0		; time of pauses for current speaker
      nsyll = 0			; number of syllables for current speaker
      nrises = 0
      nfalls = 0
      ngliss = 0
      for j from 1 to rows
         nsyll += 1
         sum_nucldur += Get value... j nucleusdur
         v = Get value... j intersyllab			; value in ST
         v_traj_inter = abs(v)
         v = Get value... j internucleusdur
         if (v < min_pause_duration)
            sum_internucldur += v
            sum_traj_inter += v_traj_inter
         else
            sum_pausedur += v
            Set value... j intersyllab 0		; needed for later
         endif
         sum_traj_intra += Get value... j trajectory
         v = Get value... j intrasyllabup		; value in ST
         dyn = 0
         if (v >= 4)
            nrises += 1
            dyn = 1
         endif
         v = Get value... j intrasyllabdown
         if (v <= -4)
            nfalls += 1
            dyn = 1
         endif
         if (dyn)					; either rise or fall, but counted once
            ngliss += 1
         endif
      endfor
      rate_j = 0 
      if (sum_nucldur + sum_internucldur > 0) 
         rate_j = nsyll / (sum_nucldur + sum_internucldur)	; speech rate for current speaker 
      endif
      traj_intra_rate = 0
      traj_inter_rate = 0
      traj_phon_rate = 0
      if (sum_nucldur > 0)
         traj_intra_rate = sum_traj_intra/sum_nucldur
      endif
      if (sum_internucldur > 0)
         traj_inter_rate = sum_traj_inter/sum_internucldur
         traj_phon_rate = (sum_traj_inter+sum_traj_intra)/(sum_internucldur+sum_nucldur)
      endif
      mean_intra_up = Get column mean (index)... intrasyllabup
      stdev_intra_up = Get column stdev (index)... intrasyllabup
      mean_abs_inter = Get column mean (index)... intersyllab	; meanwhile tweeked
      stdev_abs_inter = Get column stdev (index)... intersyllab 
      select tmptableID
      Remove
      if (ngliss > 0) 
         prises = 100*nrises/rows
         pfalls = 100*nfalls/rows
         pgliss = 100*ngliss/rows
      else
         prises = 0
         pfalls = 0
         pgliss = 0
      endif
      v1 = traj_intra_rate/stdev_of_ST		; time-normalised intrasyllabic trajectory in z-score
      v2 = traj_inter_rate/stdev_of_ST		; time-normalised intersyllabic trajectory in z-score
      v3 = traj_phon_rate/stdev_of_ST		; time-normalised total trajectory in z-score
      if (v1 == undefined)			; when insufficient nrof data
         v1 = 0
         v2 = 0
         v3 = 0
      endif
      ppauses = sum_pausedur/(sum_internucldur+sum_nucldur+sum_pausedur)
      label$ = speaker_label'speaker_j'$
      speaker_profile_'speaker_j'$ = "M_INTRA_UP='mean_intra_up:1' S_INTRA_UP='stdev_intra_up:2' " +
      ... "M_ABS_INTER='mean_abs_inter:1' S_ABS_INTER='stdev_abs_inter:2' " +
      ... "NUCL_DUR='sum_nucldur:2' INTERNUCL_DUR='sum_internucldur:3' PAUSE_DUR='sum_pausedur:3' PAUSES='ppauses:2' " +
      ... "TRAJ_INTER='sum_traj_inter:2' TRAJ_INTER_RATE='traj_inter_rate:2' TRAJ_PHON_RATE='traj_phon_rate:2'" +
      ... "TRAJ_INTRA='sum_traj_intra:2' TRAJ_INTRA_RATE='traj_intra_rate:2' " +
      ... "TRAJ_INTRA_Z_RATE='v1:3' TRAJ_INTER_Z_RATE='v2:3' TRAJ_PHON_Z_RATE='v3:3' " +
      ... "GLISS='pgliss:1' RISES='prises:1' FALLS='pfalls:1' " + "SPEECH_RATE='rate_j:3'"

   endfor ; for speaker_j
   call pitchrange_speakers_report
   call msg 'result$'
   call raw_pitchrange_speakers_report
   call msg 'result$'
   call pitchprofile_speakers_report
   call msg 'result$'
endproc


procedure pitchrange_speakers_report
   buf$ = ""
   ; Find length of longest speaker label
      len = length ("Speaker label")
      for speaker_j from 1 to speakers
         s$ = speaker_label'speaker_j'$
         len = max (len, length (s$))
      endfor
   leading$ = "     "
   buf$ = "Pitch range of speaker(s): (based on 2 stylization values per nucleus)'newline$'"
   call sprint_fs len Speaker label
   buf$ = buf$ + leading$ + result$ + ": Range, Bottom, Mean, Median, Top, MeanOfST, StdevOfST" + newline$
   for speaker_j from 1 to speakers
      label$ = speaker_label'speaker_j'$
      call sprint_fs len 'label$'
      buf$ = buf$ + leading$ + result$ + ": "
      v = extractNumber (speaker_range_'speaker_j'$, "RANGE_ST=")
      buf$ = buf$ + "'v:1'ST, "
      v = extractNumber (speaker_range_'speaker_j'$, "BOTTOM_Hz=")
      buf$ = buf$ + "'v:0'Hz "
      v = extractNumber (speaker_range_'speaker_j'$, "BOTTOM_ST=")
      buf$ = buf$ + "('v:1'ST), "
      v = extractNumber (speaker_range_'speaker_j'$, "MEAN_Hz=")
      buf$ = buf$ + "'v:0'Hz "
      v = extractNumber (speaker_range_'speaker_j'$, "MEAN_ST=")
      buf$ = buf$ + "('v:1'ST), "
      v = extractNumber (speaker_range_'speaker_j'$, "MEDIAN_Hz=")
      buf$ = buf$ + "'v:0'Hz "
      v = extractNumber (speaker_range_'speaker_j'$, "MEDIAN_ST=")
      buf$ = buf$ + "('v:1'ST), "
      v = extractNumber (speaker_range_'speaker_j'$, "TOP_Hz=")
      buf$ = buf$ + "'v:0'Hz "
      v = extractNumber (speaker_range_'speaker_j'$, "TOP_ST=")
      buf$ = buf$ + "('v:1'ST), "
      v = extractNumber (speaker_range_'speaker_j'$, "MEAN_OF_ST=")
      buf$ = buf$ + "'v:1'ST, "
      v = extractNumber (speaker_range_'speaker_j'$, "STDEV_OF_ST=")
      buf$ = buf$ + "'v:1'ST"
      buf$ = buf$ + newline$
   endfor ; for speaker_j
   result$ = buf$
endproc


procedure raw_pitchrange_speakers_report
   buf$ = ""
   ; Find length of longest speaker label
      len = length ("Speaker label")
      for speaker_j from 1 to speakers
         s$ = speaker_label'speaker_j'$
         len = max (len, length (s$))
      endfor
   leading$ = "     "
   buf$ = "Pitch range of speaker(s): (based on 2 raw F0 values per nucleus)'newline$'"
   call sprint_fs len Speaker label
   buf$ = buf$ + leading$ + result$ + ": P02, Mean, Median, P98" + newline$
   for speaker_j from 1 to speakers
      label$ = speaker_label'speaker_j'$
      call sprint_fs len 'label$'
      buf$ = buf$ + leading$ + result$ + ": "
      v = extractNumber (speaker_range_'speaker_j'$, "RF0_P02_Hz=")
      buf$ = buf$ + "'v:0'Hz, "
      v = extractNumber (speaker_range_'speaker_j'$, "RF0_MEAN_Hz=")
      buf$ = buf$ + "'v:0'Hz, "
      v = extractNumber (speaker_range_'speaker_j'$, "RF0_P50_Hz=")
      buf$ = buf$ + "'v:0'Hz, "
      v = extractNumber (speaker_range_'speaker_j'$, "RF0_P98_Hz=")
      buf$ = buf$ + "'v:0'Hz"
      buf$ = buf$ + newline$
   endfor ; for speaker_j
   result$ = buf$
endproc


procedure pitchprofile_speakers_report
   buf$ = ""
   len = 0
   for speaker_j from 1 to speakers
      len = max (len, length (speaker_label'speaker_j'$))
   endfor
   len = max (len, length ("Speaker label"))
   leading$ = "     "
   buf$ = "Pitch and duration profile of speaker(s):'newline$'"
   call sprint_fs len Speaker label
   buf$ = buf$ + leading$ + result$ + ": NuclDur, InterNuclDur, TrajIntra, TrajInter, TrajPhon, TrajIntraZ, TrajInterZ, TrajPhonZ, Gliss, Rises, Falls, SpeechRate" + newline$
   for speaker_j from 1 to speakers
      label$ = speaker_label'speaker_j'$
      call sprint_fs len 'label$'
      buf$ = buf$ + leading$ + result$ + ": "
      v = extractNumber (speaker_profile_'speaker_j'$, "NUCL_DUR=")
      buf$ = buf$ + "'v:2' s, "
      v = extractNumber (speaker_profile_'speaker_j'$, "INTERNUCL_DUR=")
      buf$ = buf$ + "'v:2' s, "
      v = extractNumber (speaker_profile_'speaker_j'$, "TRAJ_INTRA_RATE=")
      buf$ = buf$ + "'v:1' ST/s, "
      v = extractNumber (speaker_profile_'speaker_j'$, "TRAJ_INTER_RATE=")
      buf$ = buf$ + "'v:1' ST/s, "
      v = extractNumber (speaker_profile_'speaker_j'$, "TRAJ_PHON_RATE=")
      buf$ = buf$ + "'v:1' ST/s, "
      v = extractNumber (speaker_profile_'speaker_j'$, "TRAJ_INTRA_Z_RATE=")
      buf$ = buf$ + "'v:1' sd/s, "
      v = extractNumber (speaker_profile_'speaker_j'$, "TRAJ_INTER_Z_RATE=")
      buf$ = buf$ + "'v:1' sd/s, "
      v = extractNumber (speaker_profile_'speaker_j'$, "TRAJ_PHON_Z_RATE=")
      buf$ = buf$ + "'v:1' sd/s, "
      v = extractNumber (speaker_profile_'speaker_j'$, "GLISS=")
      buf$ = buf$ + "'v:1'%, "
      v = extractNumber (speaker_profile_'speaker_j'$, "RISES=")
      buf$ = buf$ + "'v:1'%, "
      v = extractNumber (speaker_profile_'speaker_j'$, "FALLS=")
      buf$ = buf$ + "'v:1'%, "
      v = extractNumber (speaker_profile_'speaker_j'$, "SPEECH_RATE=")
      buf$ = buf$ + "'v:2' syll/s"
      buf$ = buf$ + newline$
   endfor ; for speaker_j
   result$ = buf$
endproc


procedure sprint_fs .len .text$
; print string using fixed length, appending blanks if necessary
   .j = .len - length(.text$) 
   if (.j > 0)
     for .k to .j
        .text$ = .text$ + " "
     endfor
   endif
   result$ = .text$
endproc


procedure pitchrange_normalized_pitch
   current_speaker = 0					; force initialization
   select nucldatID
   .nrows = Get number of rows
   for j to .nrows
      speaker_j = Get value... j speaker_id 
      if (speaker_j <> current_speaker)
         bottomST = extractNumber (speaker_range_'speaker_j'$, "BOTTOM_ST=")
         topST = extractNumber (speaker_range_'speaker_j'$, "TOP_ST=")
         rangeST = topST - bottomST
         current_speaker = speaker_j
      endif
      v = Get value... j f0_start
      v = hertzToSemitones(v) - hertzToSemitones(1)	; convert to ST rel 1Hz
      v = 100 * (v - bottomST) / rangeST		
      ; normalize to range 0..100 (arbitrary units)
      Set value... j j_prnp_start min(100, max(0, floor(v)))
      v = Get value... j f0_end
      v = hertzToSemitones(v) - hertzToSemitones(1)	; convert to ST rel 1Hz
      v = 100 * (v - bottomST) / rangeST		
      ; normalize to range 0..100 (arbitrary units)
      Set value... j j_prnp_end min(100, max(0, floor(v)))
      .dyn = Get value... j dynamic
      .hi = Get value... j hipitch
      .hi = hertzToSemitones(.hi) - hertzToSemitones(1)
      .lo = Get value... j lopitch
      .lo = hertzToSemitones(.lo) - hertzToSemitones(1)
      if (.dyn == 0)
         v = 0
      elsif (.dyn < 0)
         v = 100 * (.lo - .hi) / rangeST
      elsif (.dyn > 0)
         v = 100 * (.hi - .lo) / rangeST
      endif
      ; v = Get value... j intrasyllabup
      ; v2 = Get value... j intrasyllabdown
      ; v = v + abs(v2)					; a value in ST
      ; v = 100 * v / rangeST		; normalize to arbitrary units
      Set value... j j_prnp_intra floor(v)
   endfor
endproc


;procedure pitch_to_zscore paramID use_ST
;; Convert Pitch object (Hz) to z-score of Hz values or of ST values
;; <use_ST>	calculate z-score on ST values (else on Hz values)
;; <result1>	mean
;; <result2>	stdev
;   select paramID
;   dataID = To Matrix
;   ncols = Get number of columns
;   ;dx = Get column distance
;   ;x1 = Get x of column... 1
;; Calculate mean of voiced points
;   sum = 0	; sum of non-zero values
;   n = 0	; nr of non-zero values
;   for col from 1 to ncols
;      value = Get value in cell... 1 col
;      if (value > 0)			; else unvoiced, pitch absent
;         if (use_ST)			; convert to ST
;            value = 12 * log2 (value)
;         endif
;         sum += value
;         n += 1
;      endif
;   endfor
;   mean = sum/n
;printline z-score: mean='mean:3' n='n' ncols='ncols'
;; Calculate stdev of voiced points
;   sumdev = 0
;   for col from 1 to ncols
;      value = Get value in cell... 1 col
;      if (value > 0)			; else unvoiced, pitch absent
;         if (use_ST)			; convert to ST
;            value = 12 * log2 (value)
;         endif
;         dev = (value - mean)
;         sumdev += (dev * dev)
;      endif
;   endfor
;   stdev = sqrt(sumdev/n)
;printline z-score: mean='mean:3' stdev='stdev:3'
;; Calculate z-scores of voiced points
;ncols = min (ncols, 1000)
;   for col from 1 to ncols
;      value = Get value in cell... 1 col
;print col='col' f='value:2' 
;      if (value > 0)
;         if (use_ST)			; convert to ST
;            value = 12 * log2 (value)
;print ST='value:2' 
;         endif
;         tmp = (value-mean)/stdev
;print z='tmp:2' 
;         Set value... 1 col (value-mean)/stdev
;      endif
;printline
;   endfor
;   result1 = mean 
;   result2 = stdev
;endproc


procedure calc_localrate table dst sizeleft sizeright
; Calculate the local speech rate in nuclei/s for a local window of (sizeleft + sizeright) seconds
; dst		column where results are stored
; sizeleft	size of window on left (in seconds)
; sizeright	size of window on right (in seconds) 
   select table
   .rows = Get number of rows
   for .row to .rows			; for each nucleus/syllable in the signal
      .x1 = Get value... .row nucleusstarttime
      .x2 = Get value... .row nucleusendtime
      .n = 0				; nrof elements in context
      tleft = 0				; total time of left context
      if (sizeleft > 0 and .row > 1)
         j = 1
         repeat				; extend left context
            if (.row - j >= 1)
               .t = Get value... .row-j nucleusstarttime
               .t2 = Get value... .row-j nucleusendtime
               if (.x1 - .t <= sizeleft)
                  .n += 1
                  tleft = .x1 - .t
               elsif (.x1 - .t2 <= sizeleft)
                  tleft = .x1 - .t2
               endif
               if (.row - j == 1)
                  j = .row		; prepare for end of loop 
               endif
            endif
            j += 1
         until (.row - j < 1 or .x1 - .t >= sizeleft)
      endif
;printline calc_localrate: x1='.x1:2' n='.n' tleft='tleft:3'
      tright = 0			; total time of right context
      if (sizeright > 0 and .row < .rows)
         j = 1
         repeat				; extend right context
            if (.row + j <= .rows)
               .t1 = Get value... .row+j nucleusstarttime
               .t = Get value... .row+j nucleusendtime
               if (.t - .x2 <= sizeright)
                  .n += 1
                  tright = .t - .x2
               elsif (.t - .x2 <= sizeright)
                  tright = .t1 - .x2
               endif
            endif
            j += 1
         until (.row + j > .rows or .t - .x2 >= sizeright) 
      endif
;printline calc_localrate: x2='.x2:2' n='.n' tright='tright:3'
;tot = tleft + .x2-.x1 + tright
;printline calc_localrate: tot='tot:2'
      Set value... .row dst (.n + 1)/(tleft + .x2-.x1 + tright)
   endfor

;   .mean = Get column mean (index)... dst
;   .stdev = Get column stdev (index)... dst
;   .max = Get value... 1 dst
;   .min = .max
;   for .row to .rows		; compute max & min 
;      .y = Get value... .row dst
;      .max = max (.y, .max)
;      .min = min (.y, .min)
;   endfor
;s$ = Get column label... dst
;printline calc_localrate: col='s$' mean='.mean:2' stdev='.stdev:2' max='.max:2' min='.min:2'
endproc
